# Generated by Django 4.2.3 on 2023-08-16 08:04

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('pragmatic', '0002_auto_20201207_1214'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='deletedobject',
            options={'default_permissions': ('add', 'change', 'delete', 'view'), 'get_latest_by': 'datetime', 'ordering': ('datetime',), 'verbose_name': 'deleted object', 'verbose_name_plural': 'deleted objects'},
        ),
        migrations.AlterField(
            model_name='deletedobject',
            name='datetime',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='datetime'),
        ),
        migrations.AlterField(
            model_name='deletedobject',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='deletedobject',
            name='object_id',
            field=models.PositiveIntegerField(verbose_name='object ID'),
        ),
        migrations.AlterField(
            model_name='deletedobject',
            name='object_str',
            field=models.CharField(max_length=300, verbose_name='object representation'),
        ),
    ]



class LockingTableQuerySetMixin(object):
    def lock(self):
        """ Lock table.

        Locks the object model table so that atomic update is possible.
        Simultaneous database access request pend until the lock is unlock()'ed.

        Note: If you need to lock multiple tables, you need to do lock them
        all in one SQL clause and this function is not enough. To avoid
        dead lock, all tables must be locked in the same order.

        If no lock mode is specified, then ACCESS EXCLUSIVE, the most restrictive mode, is used.
        Read more: https://www.postgresql.org/docs/9.4/static/sql-lock.html
        """
        from django.db import connection
        cursor = connection.cursor()
        table = self.model._meta.db_table
        cursor.execute("LOCK TABLE %s" % table)
        return cursor



def import_name(name):
    components = name.split('.')
    mod = __import__('.'.join(components[0:-1]), globals(), locals(), [components[-1]])
    return getattr(mod, components[-1])


def compress(files):
    import io
    import zipfile

    file_like_object = io.BytesIO()
    zf = zipfile.ZipFile(file_like_object, mode='w')

    try:
        for file in files:
            name = file.get('name', None)
            content = file.get('content', None)
            if name and content:
                zf.writestr(name, content, compress_type=zipfile.ZIP_DEFLATED)

    except FileNotFoundError:
        print('An error occurred during compression')
    finally:
        zf.close()
        file_like_object.seek(0)
        return file_like_object



import datetime
import inspect
import io
from itertools import groupby

import requests
from django import forms
from django.conf import settings
from django.contrib import messages
from django.contrib.admin.utils import NestedObjects
from django.contrib.auth.mixins import PermissionRequiredMixin as DjangoPermissionRequiredMixin, AccessMixin
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.core.validators import EMPTY_VALUES
from django.core.paginator import EmptyPage, Paginator
from django.db import router
from django.db.models import F, QuerySet
from django.db.models.deletion import ProtectedError
from django.http.response import HttpResponseRedirect, HttpResponse
from django.shortcuts import redirect
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.inspect import method_has_no_args
from django.utils.safestring import mark_safe
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _, gettext

from pragmatic.models import DeletedObject


class ReadOnlyFormMixin(forms.BaseForm):
    def _get_cleaner(self, field):
        def clean_field():
            return getattr(self.instance, field, None)
        return clean_field

    def __init__(self, *args, **kwargs):
        super(ReadOnlyFormMixin, self).__init__(*args, **kwargs)
        without_instance = self.read_only_without_instance if hasattr(self, "read_only_without_instance") else False
        if hasattr(self, "read_only"):
            if self.instance and self.instance.pk or without_instance:
                for field in self.read_only:
                    self.fields[field].widget.attrs['readonly'] = True
                    self.fields[field].widget.attrs['disabled'] = True
                    self.fields[field].required = False
                    setattr(self, "clean_" + field, self._get_cleaner(field))


class LoginPermissionRequiredMixin(DjangoPermissionRequiredMixin):
    raise_exception = True

    def get_permission_denied_message(self):
        """
        Override this method to override the permission_denied_message attribute.
        """
        if self.permission_denied_message:
            return self.permission_denied_message

        return self.permission_required

    def handle_no_permission(self):
        self.request.user.permission_error = self.get_permission_denied_message()

        if not self.request.user.is_authenticated:
            self.raise_exception = False

        return super().handle_no_permission()


class StaffRequiredMixin(AccessMixin):
    """
    CBV mixin which verifies that the current user is staff
    """
    raise_exception = False
    permission_denied_message = gettext('You are not authorized for this operation')

    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_staff and not request.user.is_superuser:
            return self.handle_no_permission()
        return super(StaffRequiredMixin, self).dispatch(request, *args, **kwargs)

    def handle_no_permission(self):
        if self.raise_exception:
            raise PermissionDenied(self.get_permission_denied_message())

        messages.error(self.request, self.get_permission_denied_message())
        return redirect(getattr(settings, 'LOGIN_REDIRECT_URL', '/'))


class SuperuserRequiredMixin(StaffRequiredMixin):
    """
    CBV mixin which verifies that the current user is superuser
    """
    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_superuser:
            return self.handle_no_permission()
        return super(SuperuserRequiredMixin, self).dispatch(request, *args, **kwargs)


class DeleteObjectMixin(object):
    template_name = 'confirm_delete.html'
    title = _('Delete object')
    message_success = _('Object successfully deleted.')
    message_error = _('Object could not be deleted, check if some objects are not associated with it.')
    back_url = None
    failure_url = None

    def get_back_url(self):
        try:
            return self.back_url if self.back_url else self.object.get_absolute_url()
        except:
            return self.success_url

    def get_failure_url(self):
        return self.failure_url if self.failure_url else self.get_back_url()

    def get_success_url(self):
        return self.success_url

    def delete(self, request, *args, **kwargs):
        self.object = self.get_object()
        track_deleted_objects = getattr(settings, 'PRAGMATIC_TRACK_DELETED_OBJECTS', False)

        try:
            if track_deleted_objects:
                # prepare tracking data
                content_type = ContentType.objects.get_for_model(self.object, for_concrete_model=False)
                object_id = self.object.id
                object_str = str(self.object)
                user = self.request.user

            # delete object
            self.object.delete()

            # show success message if available
            if self.message_success:
                messages.success(request, self.message_success)

            if track_deleted_objects:
                # track deleted object
                DeletedObject.objects.create(
                    content_type=content_type,
                    object_id=object_id,
                    object_str=object_str,
                    user=user if user.is_authenticated else None
                )

            return HttpResponseRedirect(self.get_success_url())
        except ProtectedError:
            if self.message_error:
                messages.error(request, self.message_error)
        return HttpResponseRedirect(self.get_failure_url())

    def get_context_data(self, **kwargs):
        context_data = super(DeleteObjectMixin, self).get_context_data(**kwargs)
        context_data['title'] = self.title
        context_data['back_url'] = self.get_back_url()
        return context_data


class CheckProtectedDeleteObjectMixin(DeleteObjectMixin):
    def dispatch(self, request, *args, **kwargs):
        using = router.db_for_write(self.get_object()._meta.model)
        collector = NestedObjects(using=using)
        collector.collect([self.get_object()])
        protected = collector.protected

        if protected:
            objects = []
            grouped_dict = {}

            for object_type, grouper in groupby(protected, key=type):
                if grouped_dict.get(object_type):
                    existing_list = grouped_dict.get(object_type)
                    for obj in list(grouper):
                        existing_list.append(obj)

                    grouped_dict[object_type] = existing_list
                else:
                    grouped_dict[object_type] = list(grouper)

            for obj_type, obj_list in grouped_dict.items():
                obj_name = _(obj_type._meta.verbose_name_plural)
                objects.append(mark_safe(
                    '%(text)s' % {'text': str(obj_name) + ' (' + str(len(obj_list)) + ')'}))

                # try:
                #     url = reverse_lazy(f'{obj_type._meta.app_label}:{obj_type._meta.model_name}_list')
                #     objects.append(mark_safe(
                #         '<a href="%(url)s" target="_blank">%(text)s</a>' % {'url': url + '?person=' + str(self.get_object().pk),
                #                                                             'text': str(obj_name) + ' (' + str(len(obj_list)) + ')'}))
                # except NoReverseMatch:
                #     objects.append(mark_safe(
                #         '%(text)s' % {'text': str(obj_name) + ' (' + str(len(obj_list)) + ')'}))

            messages.error(self.request, _('Instance cannot be deleted because of related objects: {}').format(', '.join(objects)))
            return redirect(self.get_object().get_absolute_url())

        return super().dispatch(request, *args, **kwargs)


class PickadayFormMixin(object):
    def fix_fields(self, form=None, *args, **kwargs):
        self.field_array = form.fields if form else self.fields
        for field_name in self.field_array:
            is_datefield = isinstance(self.field_array[field_name], forms.fields.DateField)
            is_datetimefield = isinstance(self.field_array[field_name], forms.fields.DateTimeField)
            if is_datefield or is_datetimefield:
                self.fix_field(field_name, *args, **kwargs)

    def fix_field(self, field_name, *args, **kwargs):
        date = None
        if field_name in self.data:
            date = self.data.get(field_name)
        elif 'initial' in kwargs and field_name in kwargs.get('initial'):
            date = kwargs.get('initial').get(field_name)
        if not date and kwargs.get('instance', None) is not None:
            instance = kwargs.get('instance')
            date = getattr(instance, field_name)
        elif not date and getattr(self, 'instance', None) is not None:
            date = getattr(self.instance, field_name)
        if date:
            if type(date) == datetime.date:
                # convert date to datetime
                date = datetime.datetime.combine(date, datetime.time(0, 0))
                # convert from naive to aware (without pytz)
                date = date.replace(tzinfo=timezone.utc)

                # convert from naive to aware (with pytz)
#                import pytz
#                date = pytz.timezone("Europe/Helsinki").localize(naive, is_dst=None)

            if type(date) == datetime.date or type(date) == datetime.datetime:
                # get date in custom format in local time
                date = timezone.localtime(date)
                date = date.strftime(settings.DATE_FORMAT)
            self.field_array[field_name].widget.attrs['data-value'] = date


class FPDFMixin(object):
    DEBUG = settings.DEBUG
    FORMAT_A4 = 'A4'
    FORMATS = {
        FORMAT_A4: {
            'width': 210,
            'height': 297
        }
    }
    format = FORMAT_A4
    ORIENTATION_PORTRAIT = 'P'
    ORIENTATION_LANDSCAPE = 'L'
    ORIENTATIONS = (
        (ORIENTATION_PORTRAIT, _('Portrait')),
        (ORIENTATION_LANDSCAPE, _('Landscape'))
    )
    orientation = ORIENTATION_PORTRAIT
    margin_left = 8
    margin_right = 8
    margin_top = 8
    margin_bottom = 8

    def render(self, **kwargs):
        # Go through keyword arguments, and either save their values to our
        # instance, or raise an error.

        try:
            from django.utils import six
        except ImportError:
            import six

        for key, value in six.iteritems(kwargs):
            setattr(self, key, value)

        self.init_sizes()
        self.init_pdf()
        self.write_pdf_content()

        self.response = HttpResponse(
            content=bytes(self.pdf.output()),
            content_type='application/octet-stream'
        )
#        self.response['Content-Type'] = 'application/octet-stream; charset=UTF-8'
        self.response['Content-Disposition'] =\
            'attachment;filename="%(filename)s"' % {
            'filename': self.get_filename()
            }
        return self.response

    def get_filename(self):
        return 'document.pdf'

    def init_sizes(self):
        # page sizes
        self.page_width = self.FORMATS[self.format]['width']
        self.page_height = self.FORMATS[self.format]['height']
        if self.orientation == 'L':
            self.page_width, self.page_height = self.page_height, self.page_width

        # content sizes
        self.content_width = self.page_width - self.margin_left - self.margin_right
        self.content_height = self.page_height - self.margin_top - self.margin_bottom

    def get_pdf_instance(self):
        from fpdf import FPDF, HTMLMixin
        class MyFPDF(FPDF, HTMLMixin):
            pass
        page_format = dict(self.FORMATS).get(self.format, None)
        if page_format is not None:
            pdf = MyFPDF(self.orientation, 'mm', (page_format['width'], page_format['height']))
        else:
            pdf = MyFPDF(self.orientation, 'mm', self.format)
        return pdf

    def init_pdf(self):
        self.pdf = self.get_pdf_instance()
        self.pdf.set_margins(self.margin_left, self.margin_top, self.margin_right)
        self.pdf.set_auto_page_break(True, margin=self.margin_bottom)
        self.pdf.add_page()

    def write_pdf_content(self):
        pass


class SafePaginator(Paginator):
    def __init__(self, object_list, per_page, orphans=0, allow_empty_first_page=True, count_only_id=False):
        super().__init__(object_list, per_page, orphans, allow_empty_first_page)
        self.count_only_id = count_only_id

    @cached_property
    def count(self):
        """Return the total number of objects, across all pages, count only id if objectlist is queryset"""
        object_list = self.object_list.only('id') if isinstance(self.object_list, QuerySet) and self.count_only_id else self.object_list
        c = getattr(object_list, 'count', None)
        if callable(c) and not inspect.isbuiltin(c) and method_has_no_args(c):
            return c()
        return len(object_list)

    def validate_number(self, number):
        try:
            return super(SafePaginator, self).validate_number(number)
        except EmptyPage:
            if number > 1:
                return self.num_pages
            else:
                raise


class DisplayListViewMixin(object):
    displays = []
    paginate_by_display = {}
    paginator_class = SafePaginator

    def dispatch(self, request, *args, **kwargs):
        self.eval_get_paginate_by(request)
        self.template_name_suffix = f'_{self.display}'
        return super().dispatch(request, *args, **kwargs)

    def eval_get_paginate_by(self, request):
        paginate_values = self.paginate_by_display.get(self.display, None)
        paginate_values = paginate_values if isinstance(paginate_values, list) else [paginate_values]
        self.paginate_by = request.GET.get('paginate_by', next(iter(paginate_values)))
        self.paginate_by = int(self.paginate_by) if self.paginate_by in map(str, paginate_values) else paginate_values[0]

    @property
    def display(self):
        display = self.request.GET.get('display', self.displays[0])
        display = display if display in self.displays else self.displays[0]
        return display

    def get_paginator(self, queryset, per_page, orphans=0, allow_empty_first_page=True, **kwargs):
        return super().get_paginator(queryset, per_page, orphans=0, allow_empty_first_page=True, count_only_id=getattr(self, 'paginator_count_only_id', False), **kwargs)

    def get_paginate_by(self, queryset):
        """
        Get the number of items to paginate by current display, or ``None`` for no pagination.
        """
        return self.paginate_by

    def get_context_data(self, *args, **kwargs):
        context_data = super().get_context_data(*args, **kwargs)
        context_data['display_modes'] = self.displays
        context_data['paginate_by_display'] = self.paginate_by_display
        context_data['paginate_by'] = self.paginate_by
        return context_data


class PaginateListViewMixin(object):
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get_paginate_by(self, queryset):
        """
        Get the number of items to paginate by current display, or ``None`` for no pagination.
        """
        return self.paginate_by

    def get_context_data(self, *args, **kwargs):
        context_data = super().get_context_data(*args, **kwargs)
        return context_data


class SortingListViewMixin(object):
    sorting_options = {}

    @property
    def sorting(self):
        first_sorting_option = next(iter(self.get_sorting_options().keys())) if len(self.get_sorting_options().keys()) > 0 else None
        sorting = self.request.GET.get('sorting', first_sorting_option)
        sorting = sorting if sorting in self.get_sorting_options() else first_sorting_option
        sorting_value = self.get_sorting_options().get(sorting)
        sorting = sorting_value[1] if isinstance(sorting_value, tuple) else sorting
        return sorting

    def get_context_data(self, *args, **kwargs):
        context_data = super().get_context_data(*args, **kwargs)
        context_data['sorting_options'] = self.get_sorting_options()
        return context_data

    def get_sorting_options(self):
        return self.sorting_options

    def get_queryset(self):
        queryset = super().get_queryset()
        return self.sort_queryset(queryset)

    def sort_queryset(self, queryset):
        if not self.sorting:
            return queryset

        if isinstance(self.sorting, list):
            return queryset.order_by(*self.sorting)

        sorting = F(self.sorting[1:]).desc(nulls_last=True) if self.sorting.startswith('-') else self.sorting
        return queryset.order_by(sorting)


class SlugMixin(object):
    MAX_SLUG_LENGTH = 150
    FORCE_SLUG_REGENERATION = True
    SLUG_FIELD = 'title'

    def save(self, **kwargs):
        if self.slug in EMPTY_VALUES or self.FORCE_SLUG_REGENERATION:
            slug_field = getattr(self, self.SLUG_FIELD)
            slug = slugify(slug_field)
            self.slug = slug
            index = 1

            # Ensure uniqueness
            while self.__class__.objects.filter(slug=self.slug).exclude(pk=self.pk).exists():
                self.slug = f'{slug}-{index}'
                index += 1

        return super().save(**kwargs)


class PdfDetailMixin(object):
    inline = True

    def dispatch(self, request, *args, **kwargs):
        response = super().dispatch(request, *args, **kwargs)

        if not getattr(response, 'is_rendered', True) and callable(getattr(response, 'render', None)):
            response.render()

        response = self.render_pdf(response.content, self.get_filename(), self.inline)
        return response

    def get_filename(self):
        return f'{self.get_object()}.pdf'

    @staticmethod
    def render_pdf(html_content, filename='output.pdf', inline=True):
        htmltopdf_api_url = getattr(settings, 'HTMLTOPDF_API_URL', None)
        printmyweb_url = getattr(settings, 'PRINTMYWEB_URL', None)
        printmyweb_token = getattr(settings, 'PRINTMYWEB_TOKEN', None)

        print_api_url = htmltopdf_api_url or printmyweb_url

        kwargs = {}
        if printmyweb_token:
            kwargs['headers'] = {
                'api-key': printmyweb_token
            }

        content_type = 'inline' if inline else 'attachment'
        pdf_response = requests.post(print_api_url, data=html_content, **kwargs)
        buffer = io.BytesIO(pdf_response.content)

        from PyPDF2 import PdfFileMerger
        pdf_merger = PdfFileMerger()
        pdf_merger.append(buffer)
        pdf_merger.addMetadata({'/Title': filename})
        pdf_merger.write(buffer)
        buffer.seek(0)

        response = HttpResponse(buffer, content_type='application/pdf')
        response['Content-Disposition'] = f'{content_type}; filename="{filename}"'
        return response



from functools import wraps
from django.contrib.auth.decorators import user_passes_test
from django.contrib.auth.models import Permission
from django.core.exceptions import PermissionDenied, ObjectDoesNotExist
from python_pragmatic.classes import get_subclasses


def permissions_required(app_label, login_url=None, raise_exception=False):
    """
    Decorator for views that checks whether a user has at least one app permission
    enabled, redirecting to the log-in page if necessary.
    If the raise_exception parameter is given the PermissionDenied exception
    is raised and app label of missing permission is stored in user instance.
    """
    def check_perms(user):
        # First check if the user has the permission (even anon users)
        if user.has_module_perms(app_label):
            return True
        # In case the 403 handler should be called raise the exception
        if raise_exception:
            user.permission_error = app_label
            raise PermissionDenied
        # As the last resort, show the login form
        return False

    return user_passes_test(check_perms, login_url=login_url)


def permission_required(perm, login_url=None, raise_exception=False):
    """
    Decorator for views that checks whether a user has a particular permission
    enabled, redirecting to the log-in page if necessary.
    If the raise_exception parameter is given the PermissionDenied exception
    is raised and missing permission is stored in user instance.
    """
    def check_perms(user):
        # First check if the user has the permission (even anon users)
        if user.has_perm(perm):
            return True
        # In case the 403 handler should be called raise the exception
        if raise_exception:
            try:
                app_label, codename = perm.split('.')
                permission = Permission.objects.get(content_type__app_label=app_label, codename=codename)
            except (ObjectDoesNotExist, ValueError):
                permission = perm
            user.permission_error = permission
            raise PermissionDenied
        # As the last resort, show the login form
        return False
    return user_passes_test(check_perms, login_url=login_url)


def receiver_subclasses(signal, sender, dispatch_uid_prefix, **kwargs):
    """
    A decorator for connecting receivers and all receiver's subclasses to signals. Used by passing in the
    signal and keyword arguments to connect::

        @receiver_subclasses(post_save, MyModel, 'mymodel_post_save')
        def signal_receiver(sender, **kwargs):
            ...

    Thanks to: http://codeblogging.net/blogs/1/14/
    """
    def _decorator(func):
        all_senders = get_subclasses(sender)
        #logging.info(all_senders)
        for snd in all_senders:
            signal.connect(func, sender=snd, dispatch_uid=dispatch_uid_prefix+'_'+snd.__name__, **kwargs)
        return func
    return _decorator


LOCK_MODES = (
    'ACCESS SHARE',
    'ROW SHARE',
    'ROW EXCLUSIVE',
    'SHARE UPDATE EXCLUSIVE',
    'SHARE',
    'SHARE ROW EXCLUSIVE',
    'EXCLUSIVE',
    'ACCESS EXCLUSIVE',
)


def require_lock(model, lock='ACCESS EXCLUSIVE'):
    """
    Decorator for PostgreSQL's table-level lock functionality

    Example:
        @transaction.commit_on_success
        @require_lock(MyModel, 'ACCESS EXCLUSIVE')
        def myview(request)
            ...

    PostgreSQL's LOCK Documentation:
    http://www.postgresql.org/docs/8.3/interactive/sql-lock.html
    """

    def require_lock_decorator(view_func):
        def wrapper(*args, **kwargs):
            if lock not in LOCK_MODES:
                raise ValueError('%s is not a PostgreSQL supported lock mode.')
            from django.db import connection
            cursor = connection.cursor()
            cursor.execute(
                'LOCK TABLE %s IN %s MODE' % (model._meta.db_table, lock)
            )
            return view_func(*args, **kwargs)
        return wrapper
    return require_lock_decorator


class Cached(object):
    def __init__(self, key, version=None, user=None, per_user=True, timeout=None):
        self.cache_key = key
        self.version = version
        self.user = user
        self.per_user = per_user
        self.timeout = timeout

    def __enter__(self):
        if self.timeout == 0:
            return None

        # read cache
        return cache.get(self.key, version=self.version)

    def __exit__(self, type, value, traceback):
        pass

    @property
    def key(self):
        if self.user and self.user.is_authenticated and self.per_user:
            return '{}:user={}'.format(self.cache_key, self.user.pk)

        return self.cache_key

    def save(self, data):
        if self.timeout != 0:
            # save to cache
            cache.set(self.key, data, version=self.version, timeout=self.timeout)

    @staticmethod
    def cache_decorator(*args, **kwargs):
        def _decorator(func):
            """
            Decorator to cache return value.
            """

            @property
            @wraps(func)
            def wrapper(self, *args, **kwargs):
                key = kwargs.get('key', None)

                if not key:
                    if hasattr(self, 'cache_key') and self.cache_key:
                        key = f'{self.cache_key}.{func.__name__}'
                    else:
                        key = func.__qualname__

                timeout = kwargs.get('timeout', 3600)
                version = kwargs.get('version', getattr(self, 'cache_version', None))
                cached = cache.get(key, version=version)

                if cached is not None:
                    return cached

                value = func(self, *args, **kwargs)
                cache.set(key, value, version=version, timeout=timeout)

                return value

            return wrapper

        return _decorator


from django.views import View
from django.views.defaults import server_error

from pragmatic.mixins import SuperuserRequiredMixin


class RaiseErrorView(SuperuserRequiredMixin, View):
    def get(self, request, *args, **kwargs):
        return server_error(request)



import decimal
from django import forms
from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.forms import CheckboxSelectMultiple
from django.utils.datastructures import MultiValueDict
from django.utils.text import capfirst
from django.core import exceptions
from django.core.exceptions import ValidationError
from django.core.validators import EMPTY_VALUES
from django.utils.encoding import smart_str
from django.utils.translation import gettext_lazy as _

from django_filters.constants import EMPTY_VALUES
from django_filters.fields import BaseRangeField, BaseCSVField
from pragmatic.widgets import SliderWidget


class AlwaysValidChoiceField(forms.ChoiceField):
    def valid_value(self, value):
        return True


class AlwaysValidMultipleChoiceField(forms.MultipleChoiceField):
    def valid_value(self, value):
        return True
    
    
class TruncatedModelChoiceField(forms.ModelChoiceField):
    def __init__(self, queryset, empty_label="---------",
                 truncate_suffix='...', truncate_chars=None,
                 required=True, widget=None, label=None, initial=None,
                 help_text=None, to_field_name=None, *args, **kwargs):

        self.truncate_chars = truncate_chars
        self.truncate_suffix = truncate_suffix

        super(TruncatedModelChoiceField, self).__init__(queryset,
            empty_label=empty_label, required=required, widget=widget, label=label, initial=initial,
            help_text=help_text, to_field_name=to_field_name, *args, **kwargs)

    def label_from_instance(self, obj):
        if self.truncate_chars:
            return smart_str(obj)[:self.truncate_chars] +\
                   (smart_str(obj)[self.truncate_chars:] and self.truncate_suffix)
        return smart_str(obj)


class RangeField(forms.Field):
    default_error_messages = {
        'invalid': _("Enter a number or range of numbers with '-' separator."),
    }

    def to_python(self, value):
        """
        Validates input value. It has to be number of range of number.
        Returns the list or range limits or None for empty values.
        """
        if value in EMPTY_VALUES:
            return None
        value = value.strip()
        #value = smart_str(value)

        #if self.localize:
        #    value = formats.sanitize_separators(value)

        try:
            value = float(value)
            start = value
            stop = value
            return start, stop
        except (ValueError, TypeError):
            pass

        try:
            start, stop = value.split('-', 1)
            start = float(start.strip())
            stop = float(stop.strip())
            return start, stop
        except (ValueError, TypeError):
            raise ValidationError(self.error_messages['invalid'])


class ChoiceArrayField(ArrayField):
    """
    A field that allows us to store an array of choices.

    Uses Django 1.9's postgres ArrayField
    and a MultipleChoiceField for its formfield.

    Usage:

        choices = ChoiceArrayField(models.CharField(max_length=...,
                                                    choices=(...,)),
                                   default=[...])
    """

    def formfield(self, **kwargs):
        defaults = {
            'form_class': forms.MultipleChoiceField,
            'choices': self.base_field.choices,
            'widget': CheckboxSelectMultiple
        }
        defaults.update(kwargs)
        # Skip our parent's formfield implementation completely as we don't
        # care for it.
        # pylint:disable=bad-super-call
        return super(ArrayField, self).formfield(**defaults)


# https://bradmontgomery.net/blog/nice-arrayfield-widgets-choices-and-chosenjs/
class ArrayFieldSelectMultiple(forms.SelectMultiple):
    """This is a Form Widget for use with a Postgres ArrayField. It implements
    a multi-select interface that can be given a set of `choices`.

    You can provide a `delimiter` keyword argument to specify the delimeter used.

    """

    def __init__(self, *args, **kwargs):
        # Accept a `delimiter` argument, and grab it (defaulting to a comma)
        self.delimiter = kwargs.pop("delimiter", ",")
        super(ArrayFieldSelectMultiple, self).__init__(*args, **kwargs)

    def render_options(self, choices, value):
        # value *should* be a list, but it might be a delimited string.
        if isinstance(value, str):  # python 2 users may need to use basestring instead of str
            value = value.split(self.delimiter)
        return super(ArrayFieldSelectMultiple, self).render_options(choices, value)

    def value_from_datadict(self, data, files, name):
        if isinstance(data, MultiValueDict):
            # Normally, we'd want a list here, which is what we get from the
            # SelectMultiple superclass, but the SimpleArrayField expects to
            # get a delimited string, so we're doing a little extra work.
            return self.delimiter.join(data.getlist(name))
        return data.get(name, None)


# New version of this snippet http://djangosnippets.org/snippets/1200/
class MultiSelectFormField(forms.MultipleChoiceField):
    widget = forms.CheckboxSelectMultiple

    def __init__(self, *args, **kwargs):
        self.max_choices = kwargs.pop('max_choices', 0)
        super(MultiSelectFormField, self).__init__(*args, **kwargs)

    def clean(self, value):
        if not value and self.required:
            raise forms.ValidationError(self.error_messages['required'])
            # if value and self.max_choices and len(value) > self.max_choices:
        #     raise forms.ValidationError('You must select a maximum of %s choice%s.'
        #             % (apnumber(self.max_choices), pluralize(self.max_choices)))
        return value


class MultiSelectField(models.Field):
    def get_internal_type(self):
        return "CharField"

    def get_choices_default(self):
        return self.get_choices(include_blank=False)

    def _get_FIELD_display(self, field):
        value = getattr(self, field.attname)
        choicedict = dict(field.choices)

    def formfield(self, **kwargs):
        # don't call super, as that overrides default widget if it has choices
        defaults = {'required': not self.blank, 'label': capfirst(self.verbose_name),
                    'help_text': self.help_text, 'choices': self.choices}
        if self.has_default():
            defaults['initial'] = self.get_default()
        defaults.update(kwargs)
        return MultiSelectFormField(**defaults)

    def get_prep_value(self, value):
        return value

    def get_db_prep_value(self, value, connection=None, prepared=False):
        if isinstance(value, basestring):
            return value
        elif isinstance(value, list):
            return ",".join(value)

    def to_python(self, value):
        if value is not None:
            return value if isinstance(value, list) else value.split(',')
        return ''

    def from_db_value(self, value, expression, connection, context):
        return value

    def contribute_to_class(self, cls, name):
        super(MultiSelectField, self).contribute_to_class(cls, name)
        if self.choices:
            func = lambda self, fieldname = name, choicedict = dict(self.choices): ",".join([choicedict.get(value, value) for value in getattr(self, fieldname)])
            setattr(cls, 'get_%s_display' % self.name, func)

    def validate(self, value, model_instance):
        arr_choices = self.get_choices_selected(self.get_choices_default())
        for opt_select in value:
            if opt_select not in arr_choices:  # the int() here is for comparing with integer choices
                raise exceptions.ValidationError(self.error_messages['invalid_choice'] % value)
        return

    def get_choices_selected(self, arr_choices=''):
        if not arr_choices:
            return False
        list = []
        for choice_selected in arr_choices:
            list.append(choice_selected[0])
        return list

    def value_to_string(self, obj):
        value = self._get_val_from_obj(obj)
        return self.get_db_prep_value(value)

    def deconstruct(self):
        """
        to support Django 1.7 migrations, see also the add_introspection_rules
        section at bottom of this file for South + earlier Django versions
        """
        name, path, args, kwargs = super(MultiSelectField, self).deconstruct()
        return name, path, args, kwargs


# needed for South compatibility

try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules(rules=[], patterns=["^pragmatic\.fields\.MultiSelectField"])
except ImportError:
    pass

# Example Model

#TYPES = ((1, 'Product'),
#(2, 'Service'),
#(3, 'Skill'),
#(4, 'Partnership')),
#(5, 'Question'),
#)
#
#class Exchange(models.Model):
#    types = MultiSelectField(max_length=250, blank=True, choices=TYPES)
#    ...
#
## Example Form
#
#class ExchangeForm(forms.Form):
#    types = MultiSelectFormField(choices=TYPES)
#    ...


class SliderField(BaseRangeField):
    widget = SliderWidget

    default_error_messages = {
        'invalid_values': _('Value should not be 0.')
    }

    def __init__(self, *args, min_value=0, max_value=100, has_range=False, step=1, show_value=True, appended_text='', show_inputs=True, **kwargs):
        self.min = min_value
        self.max = max_value
        self.has_range = has_range
        self.step = step
        self.show_value = show_value
        self.show_inputs = show_inputs
        self.appended_text = appended_text
        super().__init__(*args, **kwargs)

    def prepare_value(self, value):
        value = super().prepare_value(value)

        if value is None or len(value) == 0:
            value = [self.min, self.max] if self.has_range else self.min

        if self.has_range:
            if len(value) == 1:
                value = [self.min, value[0]]

            try:
                min = float(value[0])
            except ValueError:
                min = self.min

            try:
                max = float(value[1])
            except ValueError:
                max = self.max

            value = str([min, max])

        return value

    def to_python(self, value):
        value = super().to_python(value)

        if not isinstance(value, list):
            return value

        if len(value) == 0:
            return None

        # cast values to Decimal
        try:
            value = [decimal.Decimal(i) for i in value]
        except decimal.InvalidOperation:
            return None

        # Single value
        if len(value) == 1:
            ignore_values = [self.min, self.max] if self.has_range else [self.min]
            if value[0] in ignore_values and not self.required:
                return None
            return slice(self.min, value[0]) if self.has_range else value[0]

        # Interval
        if not self.required:
            if value[0] == self.min:
                value[0] = None
            if value[1] == self.max:
                value[1] = None

        if value[0] is None and value[1] is None:
            return None

        return_value = slice(value[0], value[1])
        return return_value

    def clean(self, value):
        if value is not None:
            value = super(BaseCSVField, self).clean(value)

        if self.required and self.min not in EMPTY_VALUES and value == self.min:
            raise forms.ValidationError(
                self.error_messages['invalid_values'],
                code='invalid_values')

        return value

    def widget_attrs(self, widget):
        attrs = super().widget_attrs(widget)

        attrs.update({
            'data-slider-min': str(self.min),
            'data-slider-max': str(self.max),
            'data-slider-step': str(self.step),
            'data-slider-show-value': str(self.show_value),
            'data-slider-value-after': self.appended_text,
            'data-slider-input': 'true' if self.show_inputs else 'false'
        })

        return attrs



from django.conf import settings
from django.http import HttpResponse
from django.template import loader
from django.template.response import SimpleTemplateResponse

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:  # Django < 1.10
    # Works perfectly for everyone using MIDDLEWARE_CLASSES
    MiddlewareMixin = object


class MaintenanceModeMiddleware(MiddlewareMixin):
    template_name = 'maintenance_mode.html'

    def process_response(self, request, response):
        # don't show maintenance mode screen if not required
        if not getattr(settings, 'MAINTENANCE_MODE', False):
            return response

        # Check if we can bypass maintenance for current user
        bypass = request.user.is_authenticated and request.user.pk in getattr(settings, 'MAINTENANCE_MODE_BYPASS_USERS', [])

        # bypass maintenance mode if staff user is logged in
        if bypass:
            return response

        # render maintenance mode screen template
        try:
            context_processors = settings.TEMPLATE_CONTEXT_PROCESSORS
        except AttributeError:
            context_processors = settings.TEMPLATES[0]['OPTIONS']['context_processors']

        if 'django.core.context_processors.request' in context_processors or \
            'django.template.context_processors.request' in context_processors:
            template = loader.get_template(self.template_name)
            return HttpResponse(template.render({}, request), status=503)
        else:
            return SimpleTemplateResponse(self.template_name, status=503).render()



from django.conf import settings
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import EmailMultiAlternatives
from django.core.validators import EMPTY_VALUES
from django.template import loader, TemplateDoesNotExist


class EmailManager(object):
    @staticmethod
    def get_recipient(to):
        return to if isinstance(to, str) else to.email

    @staticmethod
    def get_recipients(to):
        if not to:
            return None

        recipient_list = []
        if isinstance(to, list):
            for r in to:
                recipient_list.append(EmailManager.get_recipient(r))
        else:
            recipient_list.append(EmailManager.get_recipient(to))

        return recipient_list

    @staticmethod
    def send_mail(to, template_prefix, subject, data=None, attachments=[], reply_to=None, request=None):
        # template
        try:
            t = loader.get_template(f'{template_prefix}.txt')
        except TemplateDoesNotExist:
            t = None

        # HTML template
        try:
            t_html = loader.get_template(f'{template_prefix}.html')
        except TemplateDoesNotExist:
            t_html = None

        # context
        context = {
            'subject': subject,
            'request': request,
            'site': get_current_site(request),
            'settings': settings
        }

        if data:
            context.update(data)

        if not isinstance(to, list):
            context.update({'recipient': to})

        # message
        message = t.render(context) if t else ''
        html_message = t_html.render(context) if t_html else ''

        # message
        email = EmailMultiAlternatives(
            subject=subject,
            body=message,
            from_email=settings.DEFAULT_FROM_EMAIL,
            to=EmailManager.get_recipients(to),
            reply_to=EmailManager.get_recipients(reply_to)
        )

        if html_message not in EMPTY_VALUES:
            email.attach_alternative(html_message, "text/html")

        # attachments
        for attachment in attachments:
            email.attach(attachment['filename'], attachment['content'], attachment['content_type'])

        if getattr(settings, 'MAILS_QUEUE', None):
            from pragmatic.jobs import send_mail_in_background
            send_mail_in_background.delay(email)
        else:
            return email.send()



import datetime
import json
import os
import re
import urllib
from django import template
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import DateField, Count, Sum
from django.db.models.functions import TruncDay
from django.template.defaultfilters import stringfilter
from django.urls import translate_url as django_translate_url
from django.utils.html import conditional_escape
from django.utils.safestring import mark_safe
from django.utils.timezone import now

from django.utils.translation import gettext_lazy as _, gettext, override as override_language

from python_pragmatic.strings import barcode as pragmatic_barcode

register = template.Library()


@register.simple_tag(takes_context=True)
def translate_url(context, lang, *args, **kwargs):
    path = kwargs.get('path', None)
    object = kwargs.get('object', None)
    callable_object = kwargs.get('callable', None)

    if path and (object or callable_object):
        raise ValueError('"path" argument can not be used together with "object" or "callable"')

    if object and not callable_object or callable_object and not object:
        raise ValueError('Both "object" or "callable" has to be defined')

    if not path:
        # custom object
        if object:
            with override_language(lang):
                path = getattr(object, callable_object)()
                return django_translate_url(path, lang)

        # context object
        context_object = context.get('object', None)

        if context_object:
            with override_language(lang):
                get_absolute_url = getattr(context_object, "get_absolute_url", None)

                if get_absolute_url and callable(get_absolute_url):
                    return context_object.get_absolute_url()

        # URL path
        # path = context['request'].path
        request = context.get('request', None)
        path = request.path if request else f'/{lang}/'

    return django_translate_url(path, lang)


@register.filter
def get_item(value, arg):
    """Gets an attribute of an object dynamically AND recursively from a string name"""

    numeric_test = re.compile(r'^\d+$')

    try:  # for dict and  __getitem__(self, name)
        return value[arg]
    except (KeyError, TypeError, IndexError):
        pass

    if "." in str(arg):
        firstarg = str(arg).split(".")[0]
        value = get_item(value, firstarg)
        arg = ".".join(str(arg).split(".")[1:])

        return get_item(value, arg)

    if hasattr(value, str(arg)):
        return getattr(value, str(arg))

    if numeric_test.match(str(arg)) and len(value) > int(arg):
        try:
            return value[int(arg)]
        except (KeyError, IndexError):
            pass

    try:
        dict_value = dict(value)

        if isinstance(dict_value, dict) and arg in dict_value:
            return dict_value[arg]
    except ValueError:
        pass

    return getattr(settings, 'TEMPLATE_STRING_IF_INVALID', None)


@register.filter()
def get_list(querydict, param):
    return querydict.getlist(param)


@register.filter
@stringfilter
def split(string, sep):
    """Return the string split by sep.

    Example usage: {{ value|split:"/" }}
    """
    return string.split(sep)


@register.filter
def attribute(value, attr):
    return getattr(value, attr)


@register.filter
def klass(ob):
    return ob.__class__.__name__


@register.filter
def class_name(ob):
    return ob.__class__.__name__


@register.filter
def class_module(ob):
    return ob.__class__.__module__


@register.filter(name='bootstrap3_field')
def bootstrap3_field(obj):
    if 'class' in obj.field.widget.attrs:
        obj.field.widget.attrs['class'] = 'form-control ' + obj.field.widget.attrs['class']
    else:
        obj.field.widget.attrs['class'] = 'form-control'

    return obj


@register.inclusion_tag('helpers/filter_values.html', takes_context=True)
def filter_values(context, filter):
    request = context.get('request', None)

    if not request:
        return {}

    values = filtered_values(filter, request.GET)

    return {
        'full_path': request.get_full_path(),
        'filter_values': values
    }


@register.simple_tag()
def filtered_values(filter, request_data):
    values = {}

    form = filter.form
    form.full_clean()
    cleaned_data = form.cleaned_data

    for param in request_data:
        filter_name = param
        label_suffix = ''

        slice_value_name = None

        for ending in ['_before', '_after', '_min', '_max']:
            if param.endswith(ending):
                filter_name = param[:-len(ending)]

                if ending in ['_after', '_min']:
                    slice_value_name = 'start'
                    label_suffix = gettext('after') if ending == '_after' else gettext('from')
                else:
                    slice_value_name = 'stop'
                    label_suffix = gettext('before') if ending == '_before' else gettext('to')

        filter_field = filter.filters.get(filter_name, None)

        if filter_field:
            value = cleaned_data.get(filter_name, None)

            if value:
                if isinstance(value, list):
                    # multiple choice field
                    if hasattr(filter_field, 'queryset'):
                        value = ', '.join([str(v) for v in value])
                    else:
                        value_values = []

                        for v in value:
                            try:
                                v_display = dict(form.fields[filter_name].choices)[v]
                                v_display = str(v_display)
                                value_values.append(v_display)
                            except (KeyError, AttributeError):
                                try:
                                    # grouped choices
                                    choices = {}
                                    for group in form.fields[filter_name].choices:
                                        choices.update(dict(group[1]))

                                    v_display = choices[v]
                                    v_display = str(v_display)
                                    value_values.append(v_display)
                                except (KeyError, AttributeError):
                                    pass


                        value = ', '.join(value_values)
                else:
                    # choice field
                    try:
                        if not hasattr(filter_field, 'queryset'):
                            choices = form.fields[filter_name].choices

                            choices_dict = None
                            if hasattr(choices, 'choices'):
                                choices = choices.choices
                                if isinstance(value, str):
                                    choices_dict = {str(key): val for key, val in choices}

                            choices_dict = choices_dict if choices_dict is not None else dict(choices)
                            value = choices_dict[value]
                    except (KeyError, AttributeError):
                        pass

                if isinstance(value, slice):
                    # range field
                    if slice_value_name:
                        slice_value = getattr(value, slice_value_name)

                        if slice_value_name in ['start', 'stop']:
                            first_value = form.fields[filter_name].fields[0].to_python(getattr(value, 'start'))
                            last_value = form.fields[filter_name].fields[1].to_python(getattr(value, 'stop'))

                            if first_value != last_value:
                                slice_value = first_value if slice_value_name == 'start' else last_value

                        value = slice_value
                    else:
                        if value.start and value.stop:
                            value = ' - '.join([str(value.start), str(value.stop)])
                        else:
                            if value.start:
                                value = gettext('at least') + ' ' + str(value.start)
                            elif value.stop:
                                value = gettext('up to') + ' ' + str(value.stop)

                values[param] = {
                    'label': (filter_field.label + ' ' + label_suffix).strip(),
                    'value': value
                }

    return values


@register.simple_tag()
def num_applied_filters(filter, request_data):
    form = filter.form
    form.full_clean()
    cleaned_data = form.cleaned_data

    num_applied_filters = 0

    for param in request_data:
        filter_name = param

        for ending in ['_before', '_after', '_min', '_max']:
            if param.endswith(ending):
                filter_name = param[:-len(ending)]

        filter_field = filter.filters.get(filter_name, None)

        if filter_field:
            value = cleaned_data.get(filter_name, None)

            if value:
                num_applied_filters += 1

    return num_applied_filters


@register.filter('filtered_objects_counts')
def filtered_objects_counts(filtered, all):
    try:
        if filtered == all or all == 0:
            return mark_safe('%s: <strong>%d</strong>' % (gettext('total'), all))
        else:
            percent = 100 * float(filtered) / all
            return mark_safe('<strong>%d (%.2f%%)</strong> %s %d' % (filtered, percent, gettext('filtered, from a total of'), all))
    except (ValueError, TypeError):
        return ''


@register.filter
@stringfilter
def qrcode(value, alt=None):
    """
    Generate QR Code image from a string with the Google charts API

    http://code.google.com/intl/fr-FR/apis/chart/types.html#qrcodes

    Exemple usage --
    {{ my_string|qrcode:"my alt" }}

    <img src="http://chart.apis.google.com/chart?chs=150x150&amp;cht=qr&amp;chl=my_string&amp;choe=UTF-8" alt="my alt" />
    """

    url = conditional_escape("http://chart.apis.google.com/chart?%s" % \
                             urllib.urlencode({'chs': '250x250', 'cht': 'qr', 'chl': value, 'choe': 'UTF-8'}))
    alt = conditional_escape(alt or value)

    return mark_safe("""<img class="qrcode" src="%s" width="250" height="250" alt="%s" />""" % (url, alt))


@register.filter
@stringfilter
def barcode(code, args=None):
    from barcode.errors import IllegalCharacterError

    try:
        barcode = pragmatic_barcode(code, args)
        # return as HTML element
        return mark_safe('<img src="data:image/png;base64,' + barcode + '" />')
    except IllegalCharacterError as e:
        message = str(e)

        if ':' in message:
            character = message.split(':', 1)[1]
            return '{}: {}'.format(_('Invalid characters'), character)
        else:
            return message


@register.inclusion_tag('helpers/pagination.html', takes_context=True)
def paginator(context, objects, page_ident='page', anchor=None, adjacent=2):
    page_range = objects.paginator.page_range
    number = objects.number

    page_numbers = [n for n in range(number - adjacent, number + adjacent + 1) if n > 0 and n <= len(page_range)]

    show_left_dots = True
    if number - adjacent - 1 == 1:
        show_left_dots = False

    show_right_dots = True
    if number + adjacent + 1 == len(page_range):
        show_right_dots = False

    page_obj = objects.paginator.page(objects.number)

    return {
        'anchor': anchor,
        'request': context.get('request', None),
        'page_ident': page_ident,
        'results_per_page': objects.paginator.per_page,
        'page': objects.number,
        'pages': page_range,
        'count': len(page_range),
        'total_count': objects.paginator.count,
        'page_numbers': page_numbers,
        'next': objects.next_page_number,
        'previous': objects.previous_page_number,
        'has_next': objects.has_next,
        'has_previous': objects.has_previous,
        'show_first': 1 not in page_numbers,
        'show_last': False if len(page_range) - number <= adjacent else True,
        'show_left_dots': show_left_dots,
        'show_right_dots': show_right_dots,
        'start': page_obj.start_index(),
        'end': page_obj.end_index()
    }


@register.filter(is_safe=False)
def divide(value, arg):
    """Divides the value by argument."""
    try:
        return float(value) / float(arg)
    except ZeroDivisionError:
        return float(value)
    except (ValueError, TypeError):
        return ''


@register.filter(is_safe=False)
def multiply(value, arg):
    """Multiplies the value by argument."""
    try:
        return float(value) * float(arg)
    except (ValueError, TypeError):
        return ''


@register.filter(is_safe=False)
def addition(value, arg):
    """Adds the arg to the value."""
    try:
        return float(value) + float(arg)
    except (ValueError, TypeError):
        return ''


@register.filter(is_safe=False)
def subtract(value, arg):
    """Subtracts the arg to the value."""
    try:
        return float(value) - float(arg)
    except (ValueError, TypeError):
        return ''


@register.filter()
def translate(text):
    return gettext(text)


@register.filter
def filename(value):
    try:
        return os.path.basename(str(value.file.name))
    except:
        return value


@register.tag
def capture(parser, token):
    """
    Capture contents of block into context
    --------------------------------------

    Use case: variable accessing based on current variable values.

    {% capture foo %}{{ foo.value }}-suffix{% endcapture %}
    {% if foo in bar %}{% endif %}

    Created on Monday, February 2012 by Yuji Tomita
    """
    nodelist = parser.parse(('endcapture',))
    parser.delete_first_token()
    varname = token.contents.split()[1]
    return CaptureNode(nodelist, varname)


class CaptureNode(template.Node):
    def __init__(self, nodelist, varname):
        self.nodelist = nodelist
        self.varname = varname

    def render(self, context):
        context[self.varname] = self.nodelist.render(context)
        return ''


def modify_query_param(url, param, action):
    from urllib import parse
    from django.http import QueryDict

    # parse URL
    parsed_url = parse.urlparse(url)

    # path
    path = parsed_url.path

    # params as string
    params = parsed_url.query

    # convert params string to querydict
    querydict = QueryDict(params, mutable=True)

    for key, value in QueryDict(param).lists():
        if action == 'remove':
            # remove param from querydict (if exists)
            if key in querydict:
                querydict.pop(key)
        elif action == 'add':
            # add param to querydict (if not exists)
            if key not in querydict:
                querydict.setlist(key, value)
        elif action == 'replace':
            # add (and replace) param in querydict
            querydict.setlist(key, value)

    # encode params to string
    encoded_params = querydict.urlencode()

    # construct new path
    new_url = '{}?{}'.format(path, encoded_params)

    return new_url.strip('?')


@register.filter()
def add_query_param(url, param):
    return modify_query_param(url, param, action='add')


@register.filter()
def replace_query_param(url, param):
    return modify_query_param(url, param, action='replace')


@register.filter()
def remove_query_param(url, param):
    return modify_query_param(url, param, action='remove')


@register.filter
def url_anchor(html):
    pat = re.compile(r'(https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|www\.[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9]\.[^\s]{2,}|www\.[a-zA-Z0-9]\.[^\s]{2,})')
    sub = re.sub(pat, lambda x: '<a class="url-anchor" href="' + x.group(1) + '">' + x.group(1) + '</a>', html)
    return mark_safe(sub)


@register.inclusion_tag('helpers/display_modes.html', takes_context=True)
def display_modes(context):
    request = context['request']
    url = request.path
    modes = context.get('display_modes', ['list', 'table'])
    current_mode = request.GET.get('display', modes[0])
    localized_modes = {
        'table': _('Table'),
        'list': _('List'),
        'columns': _('Columns'),
        'map': _('Map'),
        'matrix': _('Matrix'),
        'tv': _('TV'),
        'graph': _('Graph'),
        'calendar': _('Calendar'),
    }
    displays = get_displays(current_mode, localized_modes, modes, request, url)

    current_paginate_by = ''
    paginate_values = []
    available_paginate_values = context.get('paginate_by_display')

    if available_paginate_values:
        available_paginate_values = available_paginate_values.get(current_mode, [None])
        available_paginate_values = available_paginate_values if isinstance(available_paginate_values, list) else [available_paginate_values]
        current_paginate_by = get_current_paginate_by(available_paginate_values, request)
        paginate_values = get_paginate_by(available_paginate_values, request, url)

    return {
        'displays': displays,
        'paginate_values': paginate_values,
        'current_paginate_by': current_paginate_by
    }


def get_current_paginate_by(available_paginate_values, request):
    current_paginate_by = request.GET.get('paginate_by', available_paginate_values[0])
    return int(current_paginate_by) if current_paginate_by in map(str, available_paginate_values) else available_paginate_values[0]


def get_displays(current_mode, localized_modes, modes, request, url):
    displays = []

    for mode in modes:
        display = {}

        params = request.GET.copy()
        params['display'] = mode

        is_active = current_mode == mode
        full_url = f'{url}?{params.urlencode()}'

        display.update({
            'mode': mode,
            'active': is_active,
            'url': full_url,
            'localized_mode': localized_modes.get(mode, mode)
        })

        displays.append(display)

    return displays


def get_paginate_by(available_paginate_values, request, url):
    paginate_by_list = []

    if len(available_paginate_values) > 1:
        for available_paginate_by in available_paginate_values:
            paginate_by = {}

            params = request.GET.copy()
            params['paginate_by'] = available_paginate_by

            full_url = f'{url}?{params.urlencode()}'

            paginate_by.update({
                'value': available_paginate_by,
                'url': full_url,
            })

            paginate_by_list.append(paginate_by)

    return paginate_by_list


@register.inclusion_tag('helpers/sorting.html', takes_context=True)
def sorting(context):
    sorting_list = []
    request = context['request']
    sorting_options = context.get('sorting_options', {'-created': 'Recently created'})
    url = request.path
    current_sorting = None

    if len(sorting_options.keys()) > 0:
        first_sorting_option = next(iter(sorting_options.keys()))
        current_sorting = request.GET.get('sorting', first_sorting_option)
        current_sorting = current_sorting if current_sorting in sorting_options else first_sorting_option
        current_sorting = sorting_options.get(current_sorting, current_sorting)
        current_sorting = str(current_sorting[0]) if isinstance(current_sorting, tuple) else current_sorting

        for sorting_option in sorting_options:
            sorting = {}

            params = request.GET.copy()
            params['sorting'] = sorting_option
            full_url = f'{url}?{params.urlencode()}'
            mode = sorting_options.get(sorting_option)

            sorting.update({
                'mode': mode[0] if isinstance(mode, tuple) else mode,
                'url': full_url,
            })

            sorting_list.append(sorting)

    return {
        'sorting_list': sorting_list,
        'current_sorting': current_sorting
    }


@register.filter
def order_by(queryset, order_by):
    return queryset.order_by(order_by)


@register.filter
def add_days(days):
    return now() + datetime.timedelta(days=days)


@register.filter
def add_months(months):
    from dateutil.relativedelta import relativedelta
    return now() + relativedelta(months=months)


@register.filter(is_safe=False)
def concat(value, arg):
    """Add the arg to the value."""
    try:
        arg = str(arg) if arg else ''
        return str(value) + arg
    except Exception:
        return str(value)


@register.inclusion_tag('admin/chart.html')
def admin_chart(objects, label=_('New data'), color='red', type='bar', date_field='created'):
    chart_data = objects \
        .annotate(date=TruncDay(date_field, output_field=DateField()))\
        .values("date")\
        .annotate(y=Count("id"))\
        .order_by("-date")

    return {
        'data': json.dumps(list(chart_data), cls=DjangoJSONEncoder),
        'label': label,
        'color': color,
        'type': type
    }


@register.simple_tag()
def objects_stats(objects, count_attr, sum_attr=None):
    kwargs = {'count': Count(count_attr)}

    if sum_attr:
        kwargs.update({'sum': Sum(sum_attr)})

    return objects\
        .values(count_attr)\
        .annotate(**kwargs)\
        .order_by('-count')


@register.filter
def date_from_isoformat(date_string):
    return datetime.date.fromisoformat(date_string)


@register.simple_tag()
def values_list(qs, attrs):
    attributes = attrs.split(',')
    return qs.values_list(*attributes, flat=True)


@register.filter()
def get_objects_by_ids(ids, model):
    ids_list = ids.split(',')
    content_type = ContentType.objects.get_by_natural_key(*model.split('.'))

    if not content_type:
        return None

    objects = content_type.model_class().objects.filter(id__in=ids_list)
    return ', '.join(str(obj) for obj in objects.all())



from django.http import JsonResponse
from django_select2.views import AutoResponseView


class AutoSlugResponseView(AutoResponseView):
    def get(self, request, *args, **kwargs):
        """
        Return a :class:`.django.http.JsonResponse`.

        Example::

            {
                'results': [
                    {
                        'text': "foo",
                        'id': "bar"
                    }
                ],
                'more': true
            }

        """
        self.widget = self.get_widget_or_404()
        self.term = kwargs.get("term", request.GET.get("term", ""))
        self.object_list = self.get_queryset()
        context = self.get_context_data()
        return JsonResponse(
            {
                "results": [
                    {"text": self.widget.label_from_instance(obj), "id": obj.slug}
                    for obj in context["object_list"]
                ],
                "more": context["page_obj"].has_next(),
            }
        )



from collections import OrderedDict

from rest_framework import routers
from rest_framework.reverse import reverse


# credit: https://newbedev.com/using-django-rest-framework-s-browsable-api-with-apiviews
class HybridRouter(routers.DefaultRouter):
    def __init__(self, *args, **kwargs):
        super(HybridRouter, self).__init__(*args, **kwargs)
        self.view_urls = []

    def add_url(self, view):
        self.view_urls.append(view)

    def get_urls(self):
        return super(HybridRouter, self).get_urls() + self.view_urls

    def get_api_root_view(self, api_urls=[]):
        original_view = super(HybridRouter, self).get_api_root_view()

        def view(request, *args, **kwargs):
            resp = original_view(request, *args, **kwargs)
            namespace = request.resolver_match.namespace
            for view_url in self.view_urls:
                name = view_url.name
                url_name = name
                if namespace:
                    url_name = namespace + ':' + url_name
                resp.data[name] = reverse(url_name,
                                          args=args,
                                          kwargs=kwargs,
                                          request=request,
                                          format=kwargs.get('format', None))
            resp.data = OrderedDict(sorted(resp.data.items()))
            return resp
        return view



from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import gettext_lazy as _


class DeletedObject(models.Model):
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField(_('object ID'))
    object_str = models.CharField(_('object representation'), max_length=300)
    user = models.ForeignKey(get_user_model(), on_delete=models.SET_NULL,
        blank=True, null=True, default=None)
    datetime = models.DateTimeField(_('datetime'), auto_now_add=True, db_index=True)

    class Meta:
        verbose_name = _('deleted object')
        verbose_name_plural = _('deleted objects')
        ordering = ('datetime',)
        get_latest_by = 'datetime'
        default_permissions = getattr(settings, 'DEFAULT_PERMISSIONS', ('add', 'change', 'delete', 'view'))

    def __str__(self):
        return self.object_str



import json
import django_filters
from django import forms
from django.apps import apps
from django.core.cache import cache
from django.core.validators import EMPTY_VALUES
from django.db.models import Max, Min, Count, Q
from django_filters.constants import EMPTY_VALUES
from pragmatic.fields import TruncatedModelChoiceField, RangeField, SliderField
from django.core.exceptions import ImproperlyConfigured


class ArrayFilter(django_filters.CharFilter):
    lookup = 'contains'
    array_size = 5

    def __init__(self, array_size=5, **kwargs):
        super(ArrayFilter, self).__init__(**kwargs)
        self.array_size = array_size

    def filter(self, qs, value):
        if value not in EMPTY_VALUES:
            kwargs = Q(**{'%s__%s' % (self.field_name, self.lookup): [value]})

            for i in range(self.array_size):
                kwargs |= Q(**{'%s__%d__%s' % (self.field_name, i, self.lookup_expr): value})

            return qs.filter(kwargs)
        return qs


class TruncatedModelChoiceFilter(django_filters.Filter):
    field_class = TruncatedModelChoiceField


class OneFieldRangeFilter(django_filters.Filter):
    field_class = RangeField

    def filter(self, qs, value):
        if value:
            return qs.filter(**{'%s__range' % self.field_name: (value[0], value[1])})
        return qs


class IntegerFilter(django_filters.NumberFilter):
    field_class = forms.IntegerField


class PositiveBooleanFilter(django_filters.BooleanFilter):
    def filter(self, qs, value):
        if not value:
            return qs
        return super().filter(qs, value)


class SliderFilter(django_filters.Filter):
    field_class = SliderField
    queryset_method = 'all'
    count = 20

    def __init__(self, *args, **kwargs):
        segment = kwargs.pop('segment', None)
        self.count = kwargs.pop('count', self.count)
        self.queryset_method = kwargs.pop('queryset_method', self.queryset_method)
        super().__init__(*args, **kwargs)

        if segment:
            self.init_segments(segment)

        self.lookup_expr = kwargs.get('lookup_expr', 'lte')

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        if isinstance(value, slice):
            if value.start:
                qs = qs.filter(**{'%s__gte' % self.field_name: value.start})

            if value.stop:
                qs = qs.filter(**{'%s__lte' % self.field_name: value.stop})
        else:
            qs = super().filter(qs, value)

        return qs

    def init_segments(self, segment):
        range_highlights = []

        dispersion = self.get_segments(segment)

        if dispersion:
            min_value = dispersion.get('min')
            max_value = dispersion.get('max')
            segments = dispersion.get('segments')

            if segments:
                max_count = max(segments.values())
                segment = (max_value - min_value) / self.count

                for num in range(0, self.count):
                    value_percent = int(segments.get("segment_" + str(num), 0) / max_count * 100)
                    class_name = 'hl-' + str(value_percent) + 'p'
                    range_highlights.append({'start': float(min_value + segment * num), 'end': float(min_value + segment * (num + 1)), 'class': class_name})

                self.field.widget.attrs.update({'data-slider-max': max_value})
                self.field.widget.attrs.update({'data-slider-min': min_value})
                self.field.widget.attrs.update({'data-slider-rangeHighlights': json.dumps(range_highlights)})
                self.field.min = min_value
                self.field.max = max_value

    def get_segment_details(self, segment):
        app_label, model_name, field_name = segment.split('.')
        model_class = apps.get_model(app_label=app_label, model_name=model_name)
        cache_key = 'slider_segments'
        cache_version = '{}_{}_{}'.format(self.count, model_class.__name__, field_name)

        return {
            'app_label': app_label,
            'model_name': model_name,
            'field_name': field_name,
            'cache_key': cache_key,
            'cache_version': cache_version,
        }

    def get_segments(self, segment):
        # read cache
        segment_details = self.get_segment_details(segment)

        try:
            saved_value = cache.get(segment_details['cache_key'], version=segment_details['cache_version'])
        except ImproperlyConfigured:
            saved_value = None

        if saved_value is not None:
            return saved_value

        model_class = apps.get_model(app_label=segment_details['app_label'], model_name=segment_details['model_name'])
        field_name = segment_details['field_name']

        qs = model_class._default_manager
        qs = getattr(qs, self.queryset_method)()

        try:
            if not qs.exists():
                return {}
        except:
            return {}

        qs = qs.values(field_name, 'id')
        min_max_count_value = qs.aggregate(Max(field_name), Min(field_name), Count('id'))
        min_value = min_max_count_value[f'{field_name}__min']
        max_value = min_max_count_value[f'{field_name}__max']
        count_values = min_max_count_value['id__count']
        aggregate_params = {}
        segments = {}

        if max_value is not None and min_value is not None:
            segment = (max_value - min_value) / self.count

            for num in range(0, self.count):
                segment_name = 'segment_' + str(num)
                query_args = {f'{field_name}__range': (min_value + segment * num, min_value + segment * (num + 1))}
                aggregate_params.update({segment_name: Count('pk', filter=Q(**query_args))})

            segments = qs.aggregate(**aggregate_params)

        store = {
            'segments': segments,
            'min': min_value,
            'max': max_value,
            'count': count_values,
        }

        # save into cache
        try:
            cache.set(segment_details['cache_key'], store, version=segment_details['cache_version'], timeout=86400)
        except ImproperlyConfigured:
            pass

        return store



import sqlparse
from debug_toolbar.panels.sql import SQLPanel as ToolbarSQLPanel
from debug_toolbar.panels.sql import views
from debug_toolbar.panels.sql.forms import SQLSelectForm
from django.http import HttpResponseBadRequest
from django.urls import re_path

try:
    # Django>=3.0
    from django.shortcuts import render
except ImportError:
    # older Django
    from django.shortcuts import render_to_response as render


from django.views.decorators.csrf import csrf_exempt


class SQLPanel(ToolbarSQLPanel):
    @classmethod
    def get_urls(cls):
        return [
            re_path(r'^sql_select/$', views.sql_select, name='sql_select'),
            re_path(r'^sql_explain/$', sql_explain, name='sql_explain'),
            re_path(r'^sql_profile/$', views.sql_profile, name='sql_profile'),
        ]


@csrf_exempt
def sql_explain(request):
    """Returns the output of the SQL EXPLAIN on the given query"""
    form = SQLSelectForm(request.POST or None)

    if form.is_valid():
        sql = form.cleaned_data['raw_sql']
        params = form.cleaned_data['params']
        vendor = form.connection.vendor
        cursor = form.cursor

        if vendor == 'sqlite':
            # SQLite's EXPLAIN dumps the low-level opcodes generated for a query;
            # EXPLAIN QUERY PLAN dumps a more human-readable summary
            # See http://www.sqlite.org/lang_explain.html for details
            cursor.execute("EXPLAIN QUERY PLAN %s" % (sql,), params)
        elif vendor == 'postgresql':
            cursor.execute("EXPLAIN ANALYZE %s" % (sql,), params)
        else:
            cursor.execute("EXPLAIN %s" % (sql,), params)

        headers = [d[0] for d in cursor.description]
        result = cursor.fetchall()
        cursor.close()
        context = {
            'result': result,
            'sql': form.reformat_sql(),
            'sql_raw': sqlparse.format(form.cleaned_data['sql'], reindent=True, keyword_case='upper'),
            'duration': form.cleaned_data['duration'],
            'headers': headers,
            'alias': form.cleaned_data['alias'],
        }
        # Using render()/render_to_response() avoids running global context processors.
        return render('panels/sql_explain.html', context)
    return HttpResponseBadRequest('Form errors')



from collections import defaultdict
from functools import wraps
from pprint import pprint

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import signals as django_signals
from django.db.models.signals import pre_init, post_init, post_save, pre_save, pre_delete, post_delete, post_migrate, \
    pre_migrate, m2m_changed
from django.utils.timezone import now


APM_DEBUG = getattr(settings, 'APM_DEBUG', False)


def add_apm_custom_context(type, value):
    from elasticapm.traces import execution_context
    transaction = execution_context.get_transaction()

    if not transaction:
        return

    if 'custom' not in transaction.context:
        transaction.context['custom'] = {}

    if type not in transaction.context['custom']:
        transaction.context['custom'][type] = [value]
    else:
        transaction.context['custom'][type].append(value)


def apm_custom_context(type, instance_attr='instance'):
    """
    A decorator for connecting functions to APM. Used by passing the context type:

        @apm_custom_context('signals')
        def my_custom_signal(sender, instance, **kwargs):
            ...

        @apm_custom_context('tasks')
        def my_custom_task(arg1, arg2):
            ...
    """
    def _decorator(func):
        """
        Decorator to send custom information to APM.
        """

        def wrapper(*args, **kwargs):
            try:
                from elasticapm.traces import execution_context

                apm_message = None

                if type == 'signals':
                    instance = kwargs.get(instance_attr, None)

                    if instance:
                        signal = kwargs.get('signal')
                        signal_name = SignalsHelper.get_signal_name(signal)
                        apm_message = f'[{signal_name}]\t{func.__module__}.{func.__qualname__}({instance.__class__.__name__}: {instance.id})'.strip()

                elif type == 'tasks':
                    # execute task with given arguments

                    arguments = str(args)
                    apm_message = f'{func.__module__}.{func.__qualname__}{arguments}'

                if apm_message:
                    if APM_DEBUG:
                        print(f'apm_message [{type}]:', apm_message)

                    add_apm_custom_context(type, apm_message)

                return func(*args, **kwargs)
            except ImportError:
                # elasticapm is not installed
                return func(*args, **kwargs)

        # return wrapper
        return wraps(func)(wrapper)  # important to preserve function signature!

    return _decorator


class SignalsHelper(object):
    @staticmethod
    def add_task_to_instance(instance, func, arguments, attr_name):
        # get existing tasks
        tasks = getattr(instance, attr_name, [])

        # prepare task
        task = (func, arguments)

        # add task to the list
        tasks.append(task)

        # save tasks into instance
        setattr(instance, attr_name, tasks)

    @staticmethod
    def add_task_and_connect(sender, instance, func, arguments, signal_type='post_save'):
        attr_name = f'{signal_type}_signal_tasks'
        receiver_name = f'{signal_type}_tasks_receiver'

        if signal_type in ['post_save', 'post_delete', 'm2m_changed']:
            signal = getattr(django_signals, signal_type)
        else:
            raise NotImplementedError()

        receiver = getattr(SignalsHelper, receiver_name)

        SignalsHelper.add_task_to_instance(instance, func, arguments, attr_name)
        signal.connect(receiver=receiver, sender=sender, weak=True)

    @staticmethod
    @apm_custom_context('signals')
    def post_save_tasks_receiver(sender, instance, **kwargs):
        SignalsHelper.execute_instance_tasks(instance, 'post_save_signal_tasks')

    @staticmethod
    @apm_custom_context('signals')
    def post_delete_tasks_receiver(sender, instance, **kwargs):
        SignalsHelper.execute_instance_tasks(instance, 'post_delete_signal_tasks')

    @staticmethod
    @apm_custom_context('signals')
    def m2m_changed_tasks_receiver(sender, instance, **kwargs):
        SignalsHelper.execute_instance_tasks(instance, 'm2m_changed_signal_tasks')

    @staticmethod
    def execute_task(task):
        # execute task with given arguments
        func = task[0]
        arguments = task[1]
        func(*arguments)

    @staticmethod
    def execute_instance_tasks(instance, attr_name):
        # start timer
        start = now()

        # get instance tasks
        tasks = getattr(instance, attr_name, [])
        total_tasks = len(tasks)

        if APM_DEBUG:
            SignalsHelper._print('>>> SignalsHelper instance tasks [{} in total]: {}'.format(total_tasks, tasks), total_tasks > 0)
        else:
            SignalsHelper._print('>>> SignalsHelper instance tasks [{} in total]'.format(total_tasks), total_tasks > 0)

        # clean instance tasks: this allows calling own save() for model instances
        setattr(instance, attr_name, [])

        for task in tasks:
            SignalsHelper.execute_task(task)

        # end timer
        duration = (now() - start).total_seconds()
        SignalsHelper._print('SignalsHelper.process_response took {} seconds'.format(duration))

    @staticmethod
    def get_db_instance(instance):
        try:
            model = type(instance)
            return model._default_manager.get(pk=instance.pk)
        except ObjectDoesNotExist:
            # object did not exist before
            return None

    @staticmethod
    def attribute_changed(instance, diff_fields, diff_contains={}, obj_exists=False):
        '''
        diff_fields: list of field names
        diff_contains: either {field_name: [vaue_1, value_2, ...]} or {field_name: {'from': [old_value_1, ...], 'to': [new_value_1, ...]}}
        '''
        obj = SignalsHelper.get_db_instance(instance)

        if not obj:
            # new object
            if obj_exists:
                return False

            return True

        # object existed before, check difference
        for field in diff_fields:
            saved_value = getattr(obj, field)
            instance_value = getattr(instance, field)

            if saved_value != instance_value:
                try:
                    # get specific values for field if supplied
                    diff_values = diff_contains[field]
                except KeyError:
                    return True

                if isinstance(diff_values, dict):
                    from_values = diff_values.get('from', [])
                    to_values = diff_values.get('to', [])

                    if from_values and to_values:
                        # from and to values provided
                        if saved_value in from_values and instance_value in to_values:
                            return True
                    elif from_values:
                        # only from values provided
                        if saved_value in from_values:
                            return True
                    elif to_values:
                        # only to values provided
                        if instance_value in to_values:
                            return True
                    else:
                        # empty dict provided
                        return True
                elif isinstance(diff_values, list):
                    if not diff_values:
                        # empty list provided
                        return True
                    elif saved_value in diff_values or instance_value in diff_values:
                        # either old or new value is in provided values
                        return True

        return False

    @staticmethod
    def get_signal_name(signal):
        return next((v for v, k in django_signals.__dict__.items() if k == signal), str(signal))

    @staticmethod
    def _print(message, force_print=False):
        if (settings.DEBUG or force_print) and getattr(settings, 'TEST_PRINT_TASKS', True):
            print(message)


class temporary_disconnect_signal:
    """ Temporarily disconnect a model from a signal """

    def __init__(self, signal, receiver, sender, dispatch_uid=None):
        self.signal = signal
        self.receiver = receiver
        self.sender = sender
        self.dispatch_uid = dispatch_uid
        self.entered_connected = False

    def __enter__(self):
        # check if receiver is connected same way as signal.disconnect
        from django.dispatch.dispatcher import _make_id

        if self.dispatch_uid:
            lookup_key = (self.dispatch_uid, _make_id(self.sender))
        else:
            lookup_key = (_make_id(self.receiver), _make_id(self.sender))

        for index in range(len(self.signal.receivers)):
            (r_key, _) = self.signal.receivers[index]
            if r_key == lookup_key:
                self.entered_connected = True
                break

        if self.entered_connected:
            self.signal.disconnect(
                receiver=self.receiver,
                sender=self.sender,
                dispatch_uid=self.dispatch_uid,
            )

    def __exit__(self, type, value, traceback):
        if self.entered_connected:
            self.signal.connect(
                receiver=self.receiver,
                sender=self.sender,
                dispatch_uid=self.dispatch_uid,
                weak=False
            )


class disable_signals:
    signals = [
            pre_init, post_init,
            pre_save, post_save,
            pre_delete, post_delete,
            pre_migrate, post_migrate,
            m2m_changed,
        ]

    def __init__(self, disabled_signals=None, enabled_signals=None, disabled_receviers=None, enabled_receivers=None):
        self.enabled_receivers = enabled_receivers
        self.disabled_receivers = disabled_receviers
        self.stashed_signals = defaultdict(list)

        if disabled_signals:
            self.disabled_signals = disable_signals
        elif enabled_signals:
            self.disabled_signals = [signal for signal in self.signals if signal not in enabled_signals]
        else:
            self.disabled_signals = self.signals

    def __enter__(self):
        for signal in self.disabled_signals:
            self.disconnect(signal)

    def __exit__(self, exc_type, exc_val, exc_tb):
        for signal in list(self.stashed_signals):
            self.reconnect(signal)

    def disconnect(self, signal):
        self.stashed_signals[signal] = signal.receivers

        if self.disabled_receivers:
            signal.receivers = [receiver for receiver in self.stashed_signals[signal] if receiver[-1]().__name__ not in self.disabled_receivers]
        elif self.enabled_receivers:
            signal.receivers = [receiver for receiver in self.stashed_signals[signal] if receiver[-1]().__name__ in self.enabled_receivers]
        else:
            signal.receivers = []

    def reconnect(self, signal):
        signal.receivers = self.stashed_signals.get(signal, [])
        del self.stashed_signals[signal]



from django.conf import settings as django_settings


def date_formats(request):
    """
    Returns a lazy 'date formats' context variables.
    """
    return {
        'DATE_FORMAT_JS': django_settings.DATE_FORMAT_JS,
        'DATE_FORMAT_TAG': django_settings.DATE_FORMAT_TAG,
        'DATE_FORMAT_FULLMONTH_TAG': django_settings.DATE_FORMAT_FULLMONTH_TAG
    }


def installed_apps(request):
    return {
        'INSTALLED_APPS': django_settings.INSTALLED_APPS
    }


def url_identifier(request):
    try:
        url_name = request.resolver_match.url_name or ''
    except AttributeError:
        return {}

    namespaces = request.resolver_match.namespaces

    return {
        'url_namespaces': namespaces,
        'url_name': url_name,
        'url_id': ':'.join(namespaces + [url_name])
    }


def settings(request):
    return {
        'settings': django_settings
    }



from itertools import chain

from django import forms
from django.forms import TextInput
from django.forms.widgets import CheckboxSelectMultiple, CheckboxInput
from django.utils.encoding import force_str
from django.utils.html import format_html
from django.utils.safestring import mark_safe


class GroupedCheckboxSelectMultiple(CheckboxSelectMultiple):
    '''
    Sample usage:

    class HomeForm(forms.Form):
        rules = forms.MultipleChoiceField(initial=('b', 'g'), choices=(
            ('a', 'a'),
            ('b', 'b'),
            ('c', 'c'),
            ('d', 'd'),
            ('e', 'e'),
            ('f', 'f'),
            ('g', 'g'),
            ('h', 'h'),

        ), widget=GroupedCheckboxSelectMultiple(attrs={
            'groups': (
                (ugettext('Group 1'), {
                    'classes': ('col-md-3', ),
                    'choices': (('a', 'a'), ('b', 'b'), ),
                }),

                (ugettext('Group 2'), {
                    'classes': ('col-md-3', ),
                    'predefined_values_on_check': 'all'|None|list(),
                    'choices': (('c', 'c'), ('d', 'd'), ),
                }),

                (ugettext('Group 3'), {
                    'classes': ('col-md-3', ),
                    'choices': (('e', 'e'), ('f', 'f'), ),
                }),

                (ugettext('Group 4'), {
                    'classes': ('col-md-3', ),
                    'choices': (('g', 'g'), ('h', 'h'), ),
                }),
            )
        }))

        def __init__(self, *args, **kwargs):
            super(HomeForm, self).__init__(*args, **kwargs)
            self.fields['rules'].label = ''
    '''

    def render(self, name, value, attrs=None, choices=()):
        if value is None: value = []
        has_id = attrs and 'id' in attrs
        final_attrs = self.build_attrs(attrs, name=name)

        # No groups in widget so render checkboxes in normal way
        if 'groups' not in final_attrs:
            output = []
            # Normalize to strings
            str_values = set([force_str(v) for v in value])
            for i, (option_value, option_label) in enumerate(chain(self.choices, choices)):
                # If an ID attribute was given, add a numeric index as a suffix,
                # so that the checkboxes don't all have the same ID attribute.
                if has_id:
                    final_attrs = dict(final_attrs, id='%s_%s' % (attrs['id'], i))
                    label_for = format_html(' for="{0}"', final_attrs['id'])
                else:
                    label_for = ''

                cb = CheckboxInput(final_attrs, check_test=lambda value: value in str_values)
                option_value = force_str(option_value)
                rendered_cb = cb.render(name, option_value)
                option_label = force_str(option_label)
                output.append('<div class="checkbox"><label%s>%s %s</label></div>' % (label_for, rendered_cb, option_label))
            return mark_safe('\n'.join(output))

        output = []
        groups = final_attrs['groups']

        if 'groups' in final_attrs:
            del final_attrs['groups']

        for group_index, group in enumerate(groups):
            str_values = set([force_str(v) for v in value])
            group_id = 'group_%i' % group_index

            result = ['<ul class="list-group">']

            has_opened_checkboxes = False
            for checkbox_index, (option_value, option_label) in enumerate(group[1]['choices']):
                if has_id:
                    final_attrs = dict(final_attrs, id='%s_%s_%s' % (attrs['id'], group_index, checkbox_index))
                    label_for = format_html(' for="{0}"', final_attrs['id'])
                else:
                    label_for = ''

                cb = CheckboxInput(final_attrs, check_test=lambda value: value in str_values)
                if cb.check_test(str(option_value)):
                    has_opened_checkboxes = True
                    break

            open_group = False
            for checkbox_index, (option_value, option_label) in enumerate(group[1]['choices']):
                if has_id:
                    final_attrs = dict(final_attrs, id='%s_%s_%s' % (attrs['id'], group_index, checkbox_index))
                    label_for = format_html(' for="{0}"', final_attrs['id'])
                else:
                    label_for = ''

                if 'on_check' in final_attrs:
                    del final_attrs['on_check']

                if has_opened_checkboxes is False and 'predefined_values_on_check' in group[1]:
                    predefined_values = group[1]['predefined_values_on_check']

                    if type(predefined_values) == str:
                        if predefined_values == 'all':
                            final_attrs['on_check'] = 'checked'
                    elif type(predefined_values) == list or type(predefined_values) == tuple:
                        if option_value in predefined_values:
                            final_attrs['on_check'] = 'checked'

                cb = CheckboxInput(final_attrs, check_test=lambda value: value in str_values)

                if cb.check_test(str(option_value)):
                    open_group = True

                option_value = force_str(option_value)
                rendered_cb = cb.render(name, option_value)
                option_label = force_str(option_label)
                result.append('<li class="list-group-item"><div class="checkbox"><label%s>%s %s</label></div></li>' % (label_for, rendered_cb, option_label))
            result.append('</ul>')

            output.append('<div class="group %(classes)s"><div class="panel panel-primary"><div class="panel-heading">%(heading)s</div>%(result)s</div></div>' % {
                'heading': '<h4 class="panel-title checkbox"><label for="%(group_id)s">%(title)s</label></h4><a href="#" class="toggler"></a>' % {
                    'group_id': group_id,
                    'title': CheckboxInput().render(group_id, open_group, attrs={
                        'id': group_id,
                    }) + group[0],
                },
                'classes': ' '.join(group[1]['classes']) if 'classes' in group[1] else '',
                'result': '\n'.join(result),
            })

        return mark_safe('\n'.join(output))


class SliderWidget(TextInput):
    template_name = 'widgets/slider_input.html'

    default_attrs = {
        'data-slider-min': '0',
        'data-slider-max': '100',
        'data-slider-step': '1',
        # 'data-slider-value': '50',
        'data-slider-tooltip': 'show',
        'data-slider-show-value': 'true',
    }

    def __init__(self, attrs=None, render_value=False):
        super().__init__(attrs)
        self.render_value = render_value
        self.attrs.update(self.default_attrs)

        if attrs:
            self.attrs.update(attrs)


try:
    import mapwidgets

    class AutocompleteGooglePointFieldWidget(mapwidgets.GooglePointFieldWidget):
        @property
        def media(self):
            js = super().media._js + ["pragmatic/js/map-widget-utils.js", ]
            return forms.Media(js=js, css=super().media._css)

        def render(self, name, value, attrs=None, renderer=None):
            return super().render(name, value, attrs)


    class AutocompleteCityWidget(TextInput):
        def __init__(self, attrs=None):
            super().__init__(attrs)
            from mapwidgets.settings import mw_settings
            self.attrs['data-google-key'] = mw_settings.GOOGLE_MAP_API_KEY

        @property
        def media(self):
            return forms.Media(js=('pragmatic/js/city-autocomplete.js',))
except ImportError:
    pass



from django_rq import job
from django.conf import settings
from django.db import connection

from rq.worker import Worker, SimpleWorker


@job(getattr(settings, 'MAILS_QUEUE', 'default'))
def send_mail_in_background(email):
    email.send()


class ConnectionClosingWorker(Worker):
    """
    RQ Worker that closes the database connection before forking.

    See also https://github.com/ui/django-rq/issues/17
    """

    def execute_job(self, job, queue):
        connection.close()
        return super().execute_job(job, queue)


class ConnectionClosingSimpleWorker(SimpleWorker):
    """
        RQ Worker that closes the database connection before forking.

        See also https://github.com/ui/django-rq/issues/17
        """

    def execute_job(self, job, queue):
        connection.close()
        return super().execute_job(job, queue)



import django_rq
from django.apps.registry import Apps, apps
from django.conf import settings
from django_rq.management.commands.rqscheduler import Command as RQSchedulerCommand


class Command(RQSchedulerCommand):
    """
    Runs regular RQ scheduler but deletes all cron jobs at first
    """
    def handle(self, *args, **options):
        self.delete_scheduled_jobs()
        self.schedule_jobs()
        super().handle(*args, **options)

    def delete_scheduled_jobs(self):
        scheduler_name = 'cron'
        print(f'Deleting scheduled jobs [scheduler name = {scheduler_name}]')
        scheduler = django_rq.get_scheduler(scheduler_name)

        # Delete any existing jobs in the scheduler when the app starts up
        for job in scheduler.get_jobs():
            job.delete()

        print('Scheduled jobs deleted.')

    def schedule_jobs(self):
        # for app in [a for a in settings.INSTALLED_APPS if not a.startswith("django.")]:
        #     print(app)

        for app in apps.get_app_configs():
            if hasattr(app, 'schedule_jobs'):
                print(f'Scheduling jobs for app {app.verbose_name} ({app.name})...')
                app.schedule_jobs()

        print('Jobs scheduled')



import logging
import traceback

from django.conf import settings
from django.core import mail
from django.utils.log import AdminEmailHandler
from django.views.debug import ExceptionReporter, get_exception_reporter_filter

# Make sure a NullHandler is available
# This was added in Python 2.7/3.2
try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

# Make sure that dictConfig is available
# This was added in Python 2.7/3.2
try:
    from logging.config import dictConfig
except ImportError:
    from django.utils.dictconfig import dictConfig

getLogger = logging.getLogger

# Ensure the creation of the Django logger
# with a null handler. This ensures we don't get any
# 'No handlers could be found for logger "django"' messages
logger = getLogger('django')
if not logger.handlers:
    logger.addHandler(NullHandler())


class AlternativeAdminEmailHandler(AdminEmailHandler):
    def emit(self, record):
        try:
            request = record.request
            subject = '%s (%s IP): %s' % (
                record.levelname,
                (request.META.get('REMOTE_ADDR') in settings.INTERNAL_IPS
                 and 'internal' or 'EXTERNAL'),
                record.getMessage()
            )
            filter = get_exception_reporter_filter(request)
            request_repr = filter.get_request_repr(request)
        except Exception:
            subject = '%s: %s' % (
                record.levelname,
                record.getMessage()
            )
            request = None
            request_repr = "Request repr() unavailable."
        subject = self.format_subject(subject)

        if record.exc_info:
            exc_info = record.exc_info
            stack_trace = '\n'.join(traceback.format_exception(*record.exc_info))
        else:
            exc_info = (None, record.getMessage(), None)
            stack_trace = 'No stack trace available'

        message = "%s\n\n%s" % (stack_trace, request_repr)
        reporter = ExceptionReporter(request, is_email=True, *exc_info)
        html_message = self.include_html and reporter.get_traceback_html() or None

        # create new connection
        connection = mail.get_connection()
        connection.password = settings.ALTERNATE_EMAIL_HOST_PASSWORD
        connection.username = settings.ALTERNATE_EMAIL_HOST_USER
        connection.host = settings.ALTERNATE_EMAIL_HOST
        connection.port = settings.ALTERNATE_EMAIL_PORT
        connection.use_tls = settings.ALTERNATE_EMAIL_USE_TLS

        mail.mail_admins(subject, message, fail_silently=True, html_message=html_message, connection=connection)



from django.contrib.contenttypes.models import ContentType

from rest_framework import serializers


class ContentTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = ContentType
        exclude = []
        fields = ['id', 'app_label', 'model']

    def to_internal_value(self, data):
        if isinstance(data, dict):
            app_label = data['app_label']
            model = data['model']
        else:
            app_label, model = data.split('.')

        content_type = ContentType.objects.get_by_natural_key(app_label, model)

        return {
            'id': content_type.id,
            'app_label': app_label,
            'model': model
        }

    def to_representation(self, instance):
        return '.'.join(instance.natural_key())


class ContentTypeNaturalField(serializers.PrimaryKeyRelatedField):
    def __init__(self, **kwargs):
        self.queryset = kwargs.pop('queryset', ContentType.objects.all())
        super().__init__(**kwargs)

    def use_pk_only_optimization(self):
        return False

    def to_internal_value(self, data):
        return ContentType.objects.get_by_natural_key(*data.split('.'))

    def to_representation(self, value):
        return '.'.join(value.natural_key())



from crispy_forms.helper import FormHelper


class SingleSubmitFormHelper(FormHelper):
    def __init__(self, form=None):
        super().__init__(form)
        self.attrs['onsubmit'] = "submit.disabled=true; return true;"



from django.db.models.expressions import F as BaseF, Value as BaseValue, Func, Expression


# https://github.com/primal100/django_postgres_extensions
class OperatorMixin(object):
    CAT = '||'
    REPLACE = '#='
    DELETE = '#-'
    KEY = '->'
    KEYTEXT = '->>'
    PATH = '#>'
    PATHTEXT = '#>>'

    def cat(self, other):
        return self._combine(other, self.CAT, False)

    def replace(self, other):
        return self._combine(other, self.REPLACE, False)

    def delete(self, other):
        return self._combine(other, self.DELETE, False)

    def key(self, other):
        return self._combine(other, self.KEY, False)

    def keytext(self, other):
        return self._combine(other, self.KEYTEXT, False)

    def path(self, other):
        return self._combine(other, self.PATH, False)

    def pathtext(self, other):
        return self._combine(other, self.PATHTEXT, False)


class F(BaseF, OperatorMixin):
    pass


class Value(BaseValue, OperatorMixin):
    def as_sql(self, compiler, connection):
        if self._output_field_or_none and any(self._output_field_or_none.get_internal_type() == fieldname for fieldname in
                                      ['ArrayField', 'MultiReferenceArrayField']):
            base_field = self._output_field_or_none.base_field
            return '%s::%s[]' % ('%s', base_field.db_type(connection)), [self.value]
        return super(Value, self).as_sql(compiler, connection)



# Generated by Django 2.2.9 on 2020-12-07 11:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pragmatic', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='deletedobject',
            name='object_str',
            field=models.CharField(max_length=300),
        ),
    ]



# Generated by Django 2.0.6 on 2019-02-20 08:19

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='DeletedObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('object_id', models.PositiveIntegerField()),
                ('object_str', models.CharField(max_length=200)),
                ('datetime', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('user', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'deleted object',
                'verbose_name_plural': 'deleted objects',
                'ordering': ('datetime',),
                'get_latest_by': 'datetime',
                'default_permissions': ('list', 'view', 'add', 'change', 'delete'),
            },
        ),
    ]



#!/usr/bin/env python
from setuptools import setup, find_packages

from pragmatic import VERSION

setup(
    name='django-pragmatic',
    version=VERSION,
    description='Pragmatic tools and utilities for Django projects',
    long_description=open('README.rst').read(),
    author='Pragmatic Mates',
    author_email='info@pragmaticmates.com',
    maintainer='Pragmatic Mates',
    maintainer_email='info@pragmaticmates.com',
    url='https://github.com/PragmaticMates/django-pragmatic',
    packages=find_packages(),
    include_package_data=True,
    install_requires=('django>=4.2', 'django-filter', 'python-pragmatic'),
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Operating System :: OS Independent',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Framework :: Django',
        'License :: OSI Approved :: BSD License',
        'Development Status :: 3 - Alpha'
    ],
    license='BSD License',
    keywords="django pragmatic tools utils",
)



from django.db.models import Aggregate, FloatField, DecimalField
from django.db.models.expressions import Func, Expression
from django.utils.functional import cached_property

from pragmatic.models.expressions import F, Value


class Round(Func):
    function = 'ROUND'
    arity = 2

    @cached_property
    def output_field(self):
        return DecimalField()


# https://github.com/primal100/django_postgres_extensions
class SimpleFunc(Func):

    def __init__(self, field, *values, **extra):
        if not isinstance(field, Expression):
            field = F(field)
            if values and not isinstance(values[0], Expression):
                values = [Value(v) for v in values]
        super(SimpleFunc, self).__init__(field, *values, **extra)


class ArrayAppend(SimpleFunc):
    function = 'ARRAY_APPEND'


class ArrayPrepend(Func):
    function = 'ARRAY_PREPEND'

    def __init__(self, value, field, **extra):
        if not isinstance(value, Expression):
            value = Value(value)
            field = F(field)
        super(ArrayPrepend, self).__init__(value, field, **extra)


class ArrayRemove(SimpleFunc):
    function = 'ARRAY_REMOVE'


class ArrayReplace(SimpleFunc):
    function = 'ARRAY_REPLACE'


class ArrayPosition(SimpleFunc):
    function = 'ARRAY_POSITION'


class ArrayPositions(SimpleFunc):
    function = 'ARRAY_POSITIONS'


class ArrayLength(SimpleFunc):
    function = 'ARRAY_LENGTH'


class Median(Aggregate):
    function = 'PERCENTILE_CONT'
    name = 'median'
    output_field = FloatField()
    template = '%(function)s(0.5) WITHIN GROUP (ORDER BY %(expressions)s)'



import os
import sys

from django.apps import apps
from django.conf import settings
from django.core.management import BaseCommand, call_command
from icecream import ic


class Command(BaseCommand):
    def add_arguments(self, parser):
        # Named (optional) arguments
        parser.add_argument('--app_prefix',
            action='store',
            dest='app_prefix',
            default=None,
            help='Import starts from line number',
            type=str
        )

    def handle(self, *args, **options):
        print('CLEANING MIGRATIONS')

        app_prefix = options.get('app_prefix', None)

        if app_prefix is None:
            exit('--app_prefix is required!')

        # check if all database migrations are applied already
        # if not is_database_synchronized():
        #     # Unapplied migrations found.
        #     # call_command('showmigrations')
        #     # call_command('migrate', '--check')
        #     print('Database is not synchronised! You should run "migrate" command at first!')
        #
        #     if input("Continue anyway? (Y)es/(N)o: ").lower() != 'y':
        #         exit('Canceled')

        # get all installed apps with the given prefix
        app_names = list(filter(lambda x: x.startswith(app_prefix), settings.INSTALLED_APPS))
        ic(app_names)

        if input("Confirm? (Y)es/(N)o: ").lower() != 'y':
            exit('Canceled')

        # get apps with migrations
        apps_with_migrations = self.get_apps_with_migrations(app_names)

        # prepare SQL delete statements
        self.sql_delete_statements(apps_with_migrations, dry_run=True)

        if input("Confirm? (Y)es/(N)o: ").lower() != 'y':
            exit('Canceled')

        self.sql_delete_statements(apps_with_migrations, dry_run=False)

        # delete app migration files
        self.migration_files(apps_with_migrations, dry_run=True)

        if input("Confirm? (Y)es/(N)o: ").lower() != 'y':
            exit('Canceled')

        self.migration_files(apps_with_migrations, dry_run=False)

        # prepare SQL statements
        self.make_migrations(apps_with_migrations, dry_run=True)

        if input("Confirm? (Y)es/(N)o: ").lower() != 'y':
            exit('Canceled')

        self.make_migrations(apps_with_migrations, dry_run=False)

        print('Done')

    def get_apps_with_migrations(self, app_names):
        apps_with_migrations = []

        for app_name in app_names:
            app = get_app_config_by_name(app_name)

            if app:
                # check if migrations folder exists
                if os.path.isdir(f'{path_to_app(app_name)}/migrations'):
                    apps_with_migrations.append(app)
                else:
                    print(f'Migrations for app "{app_name}" not available ({path_to_app(app_name)}/migrations)')
            else:
                print(f'App config for app "{app_name}" not found')

        return apps_with_migrations

    def sql_delete_statements(self, apps, dry_run):
        for app in apps:
            sql = f"delete from django_migrations where app = '{app.label}' and name != '0001_initial';"

            if dry_run:
                print(sql)
            else:
                execute_sql(sql)

    def make_migrations(self, apps, dry_run):
        for app in apps:
            if dry_run:
                print(f'makemigrations {app.label}')
            else:
                call_command('makemigrations', app.label)

    def migration_files(self, apps, dry_run):
        for app in apps:
            migrations_folder = f'{path_to_app(app.name)}/migrations'

            for filename in os.listdir(migrations_folder):
                if filename.startswith('__') or not filename.endswith('.py'):
                    continue

                file_path = os.path.join(migrations_folder, filename)

                if dry_run:
                    print(f'Remove file {file_path}')
                else:
                    try:
                        # os.unlink(file_path)
                        os.remove(file_path)
                    except Exception as e:
                        print('Failed to delete %s. Reason: %s' % (file_path, e))


def get_app_config_by_name(name):
    for app_config in apps.get_app_configs():
        if app_config.name == name:
            return app_config

    return None


def path_to_app(app_name):
    root = sys.path[0]
    app_path = '/'.join(app_name.split('.'))
    return f'{root}/{app_path}'


from django.db.migrations.executor import MigrationExecutor
from django.db import connections, DEFAULT_DB_ALIAS


def execute_sql(sql, database=DEFAULT_DB_ALIAS):
    connection = connections[database]

    with connection.cursor() as cursor:
        cursor.execute(sql)


def is_database_synchronized(database=DEFAULT_DB_ALIAS):
    connection = connections[database]
    connection.prepare_database()
    executor = MigrationExecutor(connection)
    targets = executor.loader.graph.leaf_nodes()
    return not executor.migration_plan(targets)
